#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.
#
# Autogenerated by Thrift Compiler (0.16.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#




DDL_TIME = "transient_lastDdlTime"
ACCESSTYPE_NONE = 1
ACCESSTYPE_READONLY = 2
ACCESSTYPE_WRITEONLY = 4
ACCESSTYPE_READWRITE = 8
HIVE_FILTER_FIELD_OWNER = "hive_filter_field_owner__"
HIVE_FILTER_FIELD_PARAMS = "hive_filter_field_params__"
HIVE_FILTER_FIELD_LAST_ACCESS = "hive_filter_field_last_access__"
IS_ARCHIVED = "is_archived"
ORIGINAL_LOCATION = "original_location"
IS_IMMUTABLE = "immutable"
META_TABLE_COLUMNS = "columns"
META_TABLE_COLUMN_TYPES = "columns.types"
BUCKET_FIELD_NAME = "bucket_field_name"
BUCKET_COUNT = "bucket_count"
FIELD_TO_DIMENSION = "field_to_dimension"
META_TABLE_NAME = "name"
META_TABLE_DB = "db"
META_TABLE_LOCATION = "location"
META_TABLE_SERDE = "serde"
META_TABLE_PARTITION_COLUMNS = "partition_columns"
META_TABLE_PARTITION_COLUMN_TYPES = "partition_columns.types"
FILE_INPUT_FORMAT = "file.inputformat"
FILE_OUTPUT_FORMAT = "file.outputformat"
META_TABLE_STORAGE = "storage_handler"
TABLE_IS_TRANSACTIONAL = "transactional"
TABLE_NO_AUTO_COMPACT = "no_auto_compaction"
TABLE_TRANSACTIONAL_PROPERTIES = "transactional_properties"
TABLE_BUCKETING_VERSION = "bucketing_version"
DRUID_CONFIG_PREFIX = "druid."
JDBC_CONFIG_PREFIX = "hive.sql."
TABLE_IS_CTAS = "created_with_ctas"
TABLE_IS_CTLT = "created_with_ctlt"
PARTITION_TRANSFORM_SPEC = "partition_transform_spec"
NO_CLEANUP = "no_cleanup"
CTAS_LEGACY_CONFIG = "create_table_as_external"
DEFAULT_TABLE_TYPE = "defaultTableType"
TXN_ID = "txnId"
WRITE_ID = "writeId"
